import pytest

from dataclasses import dataclass, field
from typing import Union

from hologram import JsonSchemaMixin, ValidationError
from hologram.helpers import StrEnum


class MySelector(StrEnum):
    A = "a"
    B = "b"
    C = "c"


@dataclass
class RestrictAB(JsonSchemaMixin):
    foo: MySelector = field(
        metadata={"restrict": [MySelector.A, MySelector.B]}
    )
    bar: int


@dataclass
class RestrictC(JsonSchemaMixin):
    foo: MySelector = field(metadata={"restrict": [MySelector.C]})
    baz: str


@dataclass
class HasRestricted(JsonSchemaMixin):
    thing: Union[RestrictAB, int, RestrictC]


def test_encode():
    x = HasRestricted(thing=RestrictAB(foo=MySelector.A, bar=20))
    assert x.to_dict() == {"thing": {"foo": "a", "bar": 20}}

    y = HasRestricted(thing=1000)
    assert y.to_dict() == {"thing": 1000}

    z = HasRestricted(thing=RestrictC(foo=MySelector.C, baz="hi"))
    assert z.to_dict() == {"thing": {"foo": "c", "baz": "hi"}}

    with pytest.raises(ValidationError):
        x = HasRestricted(thing=RestrictAB(foo=MySelector.C, bar=20))
        x.to_dict(validate=True)


def test_decode():
    x = HasRestricted(thing=RestrictAB(foo=MySelector.A, bar=20))
    assert (
        HasRestricted.from_dict(
            {"thing": {"foo": "a", "bar": 20}}, validate=True
        )
        == x
    )

    with pytest.raises(ValidationError):
        HasRestricted.from_dict(
            {"thing": {"foo": "c", "baz": 20}}, validate=True
        )


@dataclass
class FancyRestrictBase(JsonSchemaMixin):
    foo: MySelector


@dataclass
class FancyRestrictATrue(FancyRestrictBase):
    foo: MySelector = field(metadata={"restrict": [MySelector.A]})
    is_something: bool = field(metadata={"restrict": [True]})
    bar: str


@dataclass
class FancyRestrictAFalse(FancyRestrictBase):
    foo: MySelector = field(metadata={"restrict": [MySelector.A]})
    is_something: bool = field(metadata={"restrict": [False]})
    bar: str


@dataclass
class FancyRestrictBC(FancyRestrictBase):
    foo: MySelector = field(
        metadata={"restrict": [MySelector.B, MySelector.C]}
    )
    bar: str


@dataclass
class HasFancyRestricted(JsonSchemaMixin):
    thing: Union[FancyRestrictATrue, FancyRestrictAFalse, FancyRestrictBC]


def test_multi_symmetric():
    x = HasFancyRestricted(
        thing=FancyRestrictATrue(
            foo=MySelector.A, is_something=True, bar="a and true"
        )
    )
    x_dict = {"thing": {"foo": "a", "is_something": True, "bar": "a and true"}}
    assert x.to_dict(validate=True) == x_dict
    assert HasFancyRestricted.from_dict(x_dict, validate=True) == x

    y = HasFancyRestricted(
        thing=FancyRestrictAFalse(
            foo=MySelector.A, is_something=False, bar="a and false"
        )
    )
    y_dict = {
        "thing": {"foo": "a", "is_something": False, "bar": "a and false"}
    }
    assert y.to_dict() == y_dict
    assert HasFancyRestricted.from_dict(y_dict) == y

    z = HasFancyRestricted(
        thing=FancyRestrictBC(foo=MySelector.C, bar="c and nothing")
    )
    z_dict = {"thing": {"foo": "c", "bar": "c and nothing"}}
    assert z.to_dict() == z_dict
    assert HasFancyRestricted.from_dict(z_dict) == z

    # test the non-unioned forms, too!
    assert x.thing.to_dict() == x_dict["thing"]
    assert FancyRestrictATrue.from_dict(x_dict["thing"]) == x.thing
    # test the non-unioned forms, too!
    assert y.thing.to_dict() == y_dict["thing"]
    assert FancyRestrictAFalse.from_dict(y_dict["thing"]) == y.thing
    # test the non-unioned forms, too!
    assert z.thing.to_dict() == z_dict["thing"]
    assert FancyRestrictBC.from_dict(z_dict["thing"]) == z.thing


def test_json_schema_ok():
    schema = HasFancyRestricted.json_schema()
    assert schema["additionalProperties"] is False
    assert schema["properties"] == {
        "thing": {
            "oneOf": [
                {"$ref": "#/definitions/FancyRestrictATrue"},
                {"$ref": "#/definitions/FancyRestrictAFalse"},
                {"$ref": "#/definitions/FancyRestrictBC"},
            ]
        }
    }
    assert schema["required"] == ["thing"]

    for k in "FancyRestrictATrue", "FancyRestrictAFalse", "FancyRestrictBC":
        assert k in schema["definitions"]
        assert schema["definitions"][k]["additionalProperties"] is False
    assert schema["definitions"]["FancyRestrictATrue"]["type"] == "object"

    assert schema["definitions"]["FancyRestrictATrue"]["properties"] == {
        "foo": {"enum": ["a"], "type": "string"},
        "is_something": {"enum": [True], "type": "boolean"},
        "bar": {"type": "string"},
    }
    assert sorted(schema["definitions"]["FancyRestrictATrue"]["required"]) == [
        "bar",
        "foo",
        "is_something",
    ]
    assert schema["definitions"]["FancyRestrictAFalse"]["properties"] == {
        "foo": {"enum": ["a"], "type": "string"},
        "is_something": {"enum": [False], "type": "boolean"},
        "bar": {"type": "string"},
    }
    assert sorted(
        schema["definitions"]["FancyRestrictAFalse"]["required"]
    ) == ["bar", "foo", "is_something"]

    schema["definitions"]["FancyRestrictBC"]["properties"]["foo"][
        "enum"
    ].sort()
    assert schema["definitions"]["FancyRestrictBC"]["properties"] == {
        "foo": {"enum": ["b", "c"], "type": "string"},
        "bar": {"type": "string"},
    }
    assert sorted(schema["definitions"]["FancyRestrictBC"]["required"]) == [
        "bar",
        "foo",
    ]


class OtherSelector(StrEnum):
    X = "x"
    Y = "y"
    Z = "z"


@dataclass
class InvalidRestrictedType(JsonSchemaMixin):
    foo: MySelector = field(metadata={"restrict": [MySelector.A, int]})


def test_bad_restrictions():
    with pytest.raises(ValidationError):
        InvalidRestrictedType.json_schema()


@dataclass
class FancyRestrictATrueExtras(FancyRestrictBase):
    foo: MySelector = field(metadata={"restrict": [MySelector.A]})
    is_something: bool = field(metadata={"restrict": [True]})
    bar: str
    baz: str


@dataclass
class HasFancyRestrictedOverlapping(JsonSchemaMixin):
    thing: Union[
        FancyRestrictATrue,
        FancyRestrictAFalse,
        FancyRestrictBC,
        FancyRestrictATrueExtras,
    ]


@dataclass
class HasFancyRestrictedOverlappingReversed(JsonSchemaMixin):
    thing: Union[
        FancyRestrictATrueExtras,
        FancyRestrictBC,
        FancyRestrictAFalse,
        FancyRestrictATrue,
    ]


def test_multi_overlapping_restricted():
    classes = [
        HasFancyRestrictedOverlapping,
        HasFancyRestrictedOverlappingReversed,
    ]
    for cls in classes:
        fancy_a = FancyRestrictATrue(
            foo=MySelector.A, is_something=True, bar="hello"
        )
        fancy_a_extra = FancyRestrictATrueExtras(
            foo=MySelector.A, is_something=True, bar="hello", baz="goodbye"
        )
        x = cls(thing=fancy_a)
        y = cls(thing=fancy_a_extra)

        assert cls.from_dict(x.to_dict()) == x
        assert cls.from_dict(y.to_dict()) == y



import pytest

from dataclasses import dataclass, field
from typing import Union, NewType, Optional

from hologram import JsonSchemaMixin, ValidationError
from hologram.helpers import StrEnum


class MySelector(StrEnum):
    A = "a"
    B = "b"
    C = "c"


@dataclass
class RestrictAB(JsonSchemaMixin):
    foo: MySelector = field(
        metadata={"restrict": [MySelector.A, MySelector.B]}
    )
    bar: int


@dataclass
class RestrictC(JsonSchemaMixin):
    foo: MySelector = field(metadata={"restrict": [MySelector.C]})
    baz: str


@dataclass
class A(JsonSchemaMixin):
    baz: str


@dataclass
class B(JsonSchemaMixin):
    baz: str


@dataclass
class HasRestricted(JsonSchemaMixin):
    thing: Union[A, Optional[Union[RestrictAB, RestrictC]], B]


def test_encode():
    x = HasRestricted(thing=RestrictAB(foo=MySelector.A, bar=20))
    assert x.to_dict() == {"thing": {"foo": "a", "bar": 20}}

    y = HasRestricted(thing=1000)
    assert y.to_dict() == {"thing": 1000}

    z = HasRestricted(thing=RestrictC(foo=MySelector.C, baz="hi"))
    assert z.to_dict() == {"thing": {"foo": "c", "baz": "hi"}}

    with pytest.raises(ValidationError):
        x = HasRestricted(thing=RestrictAB(foo=MySelector.C, bar=20))
        x.to_dict(validate=True)


def test_decode():
    x = HasRestricted(thing=RestrictAB(foo=MySelector.A, bar=20))
    assert (
        HasRestricted.from_dict(
            {"thing": {"foo": "a", "bar": 20}}, validate=True
        )
        == x
    )

    with pytest.raises(ValidationError):
        HasRestricted.from_dict(
            {"thing": {"foo": "c", "baz": 20}}, validate=True
        )


@dataclass
class IHaveExtremelyAnnoyingUnions(JsonSchemaMixin):
    my_field: Union[Optional[str], bool, float] = None


def test_evil_union():
    pairs = [
        (
            IHaveExtremelyAnnoyingUnions(my_field=True).to_dict(),
            {"my_field": True},
        ),
        (
            IHaveExtremelyAnnoyingUnions(my_field="1").to_dict(),
            {"my_field": "1"},
        ),
        (
            IHaveExtremelyAnnoyingUnions(my_field=1.0).to_dict(),
            {"my_field": 1.0},
        ),
        (IHaveExtremelyAnnoyingUnions().to_dict(), {}),
    ]
    for a, b in pairs:
        assert a == b
        assert IHaveExtremelyAnnoyingUnions.from_dict(b).to_dict() == a



import pytest

from dataclasses import dataclass

from hologram import ValidationError, JsonSchemaMixin
from hologram.helpers import ExtensibleJsonSchemaMixin


@dataclass
class Extensible(ExtensibleJsonSchemaMixin):
    a: str
    b: int


@dataclass
class Inextensible(JsonSchemaMixin):
    a: str
    b: int


@dataclass
class ContainsExtended(JsonSchemaMixin):
    ext: Extensible


@pytest.fixture
def extensible_dict():
    return {"a": "one", "b": 1}


@pytest.fixture
def extra_data(extensible_dict):
    extra = dict(extensible_dict)
    extra["c"] = []
    return extra


@pytest.fixture
def extensible():
    return Extensible(a="one", b=1)


@pytest.fixture
def inextensible():
    return Inextensible(a="one", b=1)


@pytest.fixture
def contains_extra_data(extra_data):
    return {"ext": dict(extra_data)}


@pytest.fixture
def contains():
    return ContainsExtended(ext=Extensible(a="one", b=1))


@pytest.fixture
def contains_dict(extensible_dict):
    return {"ext": dict(extensible_dict)}


def test_extensible(extra_data, extensible_dict, extensible):
    assert Extensible.from_dict(extra_data) == extensible
    assert Extensible.from_dict(extensible_dict) == extensible
    assert extensible.to_dict() == extensible_dict
    assert Extensible.from_dict(extra_data).to_dict() == extensible_dict


def test_inextensible(extra_data, extensible_dict, inextensible):
    assert Inextensible.from_dict(extensible_dict) == inextensible
    assert inextensible.to_dict() == extensible_dict
    with pytest.raises(ValidationError):
        Inextensible.from_dict(extra_data)


def test_contains(contains_extra_data, contains_dict, contains):
    assert ContainsExtended.from_dict(contains_dict) == contains
    assert (
        ContainsExtended.from_dict(contains_extra_data).to_dict()
        == contains_dict
    )



import pytest

from dataclasses import dataclass
from hologram import JsonSchemaMixin
from typing import Dict, Union


@dataclass
class DictFieldValue(JsonSchemaMixin):
    z: str


@dataclass
class SecondDictFieldValue(JsonSchemaMixin):
    z: int


@dataclass
class HasDictFields(JsonSchemaMixin):
    a: str
    x: Dict[str, str]
    z: Dict[str, Union[DictFieldValue, SecondDictFieldValue]]


def test_schema():
    schema = HasDictFields.json_schema()

    assert schema["type"] == "object"
    assert schema["required"] == ["a", "x", "z"]
    assert schema["properties"] == {
        "a": {"type": "string"},
        "x": {"type": "object", "additionalProperties": {"type": "string"}},
        "z": {
            "type": "object",
            "additionalProperties": {
                "oneOf": [
                    {"$ref": "#/definitions/DictFieldValue"},
                    {"$ref": "#/definitions/SecondDictFieldValue"},
                ]
            },
        },
    }
